/* Copyright (C) 2006 Imperial College London and others.

 Please see the AUTHORS file in the main source directory for a full list
 of copyright holders.

 Dr Gerard J Gorman
 Applied Modelling and Computation Group
 Department of Earth Science and Engineering
 Imperial College London

 g.gorman@imperial.ac.uk

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 USA
*/
#include <confdefs.h>

#include <mpi.h>
#include <cassert>
#include <cstdlib>
#include <vector>
#include <string>
#include <typeinfo>
#include <iostream>

#include "c++debug.h"

#define TAG_33 33
#define TAG_44 44

// In-place send-receive.
template<class T>
void allSendRecv(std::vector< std::vector<T> >& inout){
  
  // This line is important as it ensures that successive calls
  // don't overlap
  MPI::COMM_WORLD.Barrier();
  
  unsigned MyRank;
  {
    int rank = MPI::COMM_WORLD.Get_rank();
    if(rank<0){
      ERROR( mpiErrorMessage(rank) );
    MPI::COMM_WORLD.Abort(MPI_ERR_LASTCODE);
    }
    MyRank = rank;
  }

  unsigned NProcs;
  {
    int size = MPI::COMM_WORLD.Get_size();
    if(size<0){
      ERROR( mpiErrorMessage(size) );
      MPI::COMM_WORLD.Abort(MPI_ERR_LASTCODE);
    }
    NProcs = size;
  }

  // Determine the MPI datatype on the fly.
  MPI::Datatype mpi_type;
  if(typeid(T) == typeid(signed char)){
    mpi_type =  MPI::CHAR;
    ECHO("MPI::CHAR");
  }else if(typeid(T) == typeid(char)){
    mpi_type =  MPI::CHAR;
    ECHO("MPI::CHAR");
  }else if(typeid(T) == typeid(signed short int)){
    mpi_type =  MPI::SHORT;
    ECHO("MPI::SHORT");
  }else if(typeid(T) == typeid(signed int)){
    mpi_type =  MPI::INT;
    ECHO("MPI::INT");
  }else if(typeid(T) == typeid(int)){
    mpi_type =  MPI::INT;
    ECHO("MPI::INT");
  }else if(typeid(T) == typeid(signed long int)){
    mpi_type =  MPI::LONG;
    ECHO("MPI::LONG");
  }else if(typeid(T) == typeid(long)){
    mpi_type =  MPI::LONG;
    ECHO("MPI::LONG");
  }else if(typeid(T) == typeid(unsigned char)){
    mpi_type =  MPI::UNSIGNED_CHAR;
    ECHO("MPI::UNSIGNED_CHAR");
  }else if(typeid(T) == typeid(unsigned short int)){
    mpi_type =  MPI::UNSIGNED_SHORT;
    ECHO("MPI::UNSIGNED_SHORT");
  }else if(typeid(T) == typeid(unsigned int)){
    mpi_type =  MPI::UNSIGNED;
    ECHO("MPI::UNSIGNED");
  }else if(typeid(T) == typeid(unsigned)){
    mpi_type =  MPI::UNSIGNED;
    ECHO("MPI::UNSIGNED");
  }else if(typeid(T) == typeid(unsigned long int)){
    mpi_type =  MPI::UNSIGNED_LONG;
    ECHO("MPI::UNSIGNED_LONG");
  }else if(typeid(T) == typeid(float)){
    mpi_type =  MPI::FLOAT;
    ECHO("MPI::FLOAT");
  }else if(typeid(T) == typeid(double)){
    mpi_type =  MPI::DOUBLE;
    ECHO("MPI::DOUBLE");
  }else if(typeid(T) == typeid(long double)){
    mpi_type =  MPI::LONG_DOUBLE;
    ECHO("MPI::LONG_DOUBLE");
  }else{
    std::cerr << "ERROR: illegal type " << typeid(T).name()
	 << " passed into allSendRecv(std::vector< std::vector<T> >& inout)"
	 << std::endl;
    MPI::COMM_WORLD.Abort(-1);
  }
  const MPI::Datatype const_mpi_type = mpi_type;

  
  // Setup some receiving buffer space.
  std::vector< std::vector<T> > recvBuffer( NProcs );
  
  // MPI_Status sendStatus, recvStatus;
  std::vector<MPI::Request> sendRequest( NProcs );
  std::vector<MPI::Request> recvRequest( NProcs );
  
  // Send
  for(unsigned p=0; p < NProcs; p++){
    if(p == MyRank){ // Sending nothing to ourselves.
      sendRequest[ p ] = MPI::REQUEST_NULL;
    }else{
      ECHO("MPI Packet of " << inout[p].size() << " being sent to " << p);
      int count = inout[p].size();
      if(count>0){
        sendRequest[p] = MPI::COMM_WORLD.Isend(&(inout[p][0]), count, const_mpi_type, p, TAG_33);
      }else{
	sendRequest[p] = MPI::COMM_WORLD.Isend(NULL, 0, const_mpi_type, p, TAG_33);
      }
    }
  }
  
  // Receive
  recvRequest[MyRank] = MPI::REQUEST_NULL;    
  for(unsigned p=0; p<NProcs-1; p++){ // -1 because we'll not be receiving from ourselves
    MPI::Status recvStatus;

    // blocking probe for incoming message
    MPI::COMM_WORLD.Probe(MPI::ANY_SOURCE, TAG_33, recvStatus);  
    if(recvStatus.Get_error()){
      ERROR( mpiErrorMessage(recvStatus.Get_error()) );
      MPI::COMM_WORLD.Abort(MPI_ERR_LASTCODE);
    }

    const int source = recvStatus.Get_source();
    if(source<0){
      ERROR( mpiErrorMessage(source) );
      MPI::COMM_WORLD.Abort(MPI_ERR_LASTCODE);
    }

    const int count = recvStatus.Get_count(const_mpi_type);
    if(count<0){
      ERROR(mpiErrorMessage(count));
      ERROR("Receiving from "<<source);
      MPI::COMM_WORLD.Abort(MPI_ERR_LASTCODE);
    }
    
    // Allocate receive space.
    ECHO("Receiving message length of "<<count<<" data from "<<source);
    recvBuffer[source].resize(count);
    
    // Initiate non-blocking receive.
    if(count>0){
      T *buffer = &(recvBuffer[source][0]);
      recvRequest[source] = MPI::COMM_WORLD.Irecv(buffer, count, const_mpi_type, source, TAG_33);
    }else{
      recvRequest[source] = MPI::COMM_WORLD.Irecv(NULL, 0, const_mpi_type, source, TAG_33);
    }
  }
  
  MPI::Request::Waitall(NProcs, &(sendRequest[0])); // Wait all sends
  sendRequest.clear();
  inout.clear();

  MPI::Request::Waitall(NProcs, &(recvRequest[0])); // Wait all receives
  recvRequest.clear();
  
  inout.swap(recvBuffer);

  ECHO("All data received");
 
  return;
}

// In-place send-receive.
template<class T>
void allSendRecv(std::vector< std::set<T> >& inout){
  unsigned len = inout.size();

  // Write to vectors
  std::vector< std::vector<T> > in(len);
  for(unsigned i=0; i<len; i++){
    in[i].resize( inout[i].size() );
    unsigned j=0;
    for(typename std::set<T>::const_iterator it=inout[i].begin(); it!=inout[i].end(); ++it){
      in[i][j++] = *it;
    }
    inout[i].clear();
  }
  
  // Communicate
  allSendRecv(in);
  
  // Write to sets
  for(unsigned i=0; i<len; i++){
    for(typename std::vector<T>::const_iterator it=in[i].begin(); it!=in[i].end(); ++it){
      inout[i].insert(*it);
    }
    in[i].clear();
  }
  
  return;
}

template<class T>
void allSendRecv(const std::vector< std::vector<T> >& sendBuffer, 
		 std::vector< std::vector<T> >& recvBuffer){
  
  // This line is important as it ensures that successive calls
  // don't overlap
  MPI::COMM_WORLD.Barrier();

  unsigned MyRank = MPI::COMM_WORLD.Get_rank();
  unsigned NProcs = sendBuffer.size();
  assert(NProcs == sendBuffer.size());
  
  // Determine the MPI datatype on the fly.
  MPI::Datatype mpi_type;

  if(typeid(T) == typeid(signed char)){
    mpi_type =  MPI::CHAR;
  }else if(typeid(T) == typeid(char)){
    mpi_type =  MPI::CHAR;
  }else if(typeid(T) == typeid(signed short int)){
    mpi_type =  MPI::SHORT;
  }else if(typeid(T) == typeid(signed int)){
    mpi_type =  MPI::INT;
  }else if(typeid(T) == typeid(int)){
    mpi_type =  MPI::INT;
  }else if(typeid(T) == typeid(signed long int)){
    mpi_type =  MPI::LONG;
  }else if(typeid(T) == typeid(long)){
    mpi_type =  MPI::LONG;
  }else if(typeid(T) == typeid(unsigned char)){
    mpi_type =  MPI::UNSIGNED_CHAR;
  }else if(typeid(T) == typeid(unsigned short int)){
    mpi_type =  MPI::UNSIGNED_SHORT;
  }else if(typeid(T) == typeid(unsigned int)){
    mpi_type =  MPI::UNSIGNED;
  }else if(typeid(T) == typeid(unsigned)){
    mpi_type =  MPI::UNSIGNED;
  }else if(typeid(T) == typeid(unsigned long int)){
    mpi_type =  MPI::UNSIGNED_LONG;
  }else if(typeid(T) == typeid(float)){
    mpi_type =  MPI::FLOAT;
  }else if(typeid(T) == typeid(double)){
    mpi_type =  MPI::DOUBLE;
  }else if(typeid(T) == typeid(long double)){
    mpi_type =  MPI::LONG_DOUBLE;
  }else{
    std::cerr << "ERROR: illegal type " << typeid(T).name()
	 << " passed into allSendRecv(std::vector< std::vector<T> >& inout)"
	 << std::endl;
    exit(-1);
  }  
  
  // Just in case...
  recvBuffer.clear();
  recvBuffer.resize(NProcs);
  
  // MPI_Status sendStatus, recvStatus;
  std::vector<MPI::Request> sendRequest( NProcs );
  std::vector<MPI::Request> recvRequest( NProcs );
  MPI::Status recvStatus;
  
  // Set up all the non-blocking sends.
  for(unsigned p=0; p < NProcs; p++){
    if(p == MyRank){ // Sending nothing to ourselves.
      sendRequest[ p ] = MPI::REQUEST_NULL;
      continue;
    }
    
    sendRequest[p] = MPI::COMM_WORLD.Isend(&(sendBuffer[p][0]), 
					   sendBuffer[p].size(), 
					   mpi_type, p, TAG_44);
  }
  
  // Set up all the receives. These are set up with the help of an initial probe.
  // Note -1 because we'll not be receiving from ourselves.
  recvRequest[ MyRank ] = MPI::REQUEST_NULL;    
  for(unsigned p=0; p < NProcs-1; p++){ 
    
    // Blocking probe for any incoming message.
    MPI::COMM_WORLD.Probe(MPI::ANY_SOURCE, TAG_44, recvStatus);  
    int mesg_size = recvStatus.Get_count(mpi_type);
    int rank      = recvStatus.Get_source();

    // Allocate receive space.
    recvBuffer[rank].resize(mesg_size);
    
    // initiate non-blocking receive
    recvRequest[rank] = MPI::COMM_WORLD.Irecv(&(recvBuffer[rank][0]), mesg_size, mpi_type, rank, TAG_44);
  }
  
  // Wait for all sends to complete.
  MPI::Request::Waitall(NProcs, &(sendRequest[0]));
  // Wait for all receives to complete.
  MPI::Request::Waitall(NProcs, &(recvRequest[0]));
  
  return;
}

template<class T>
void allSendRecv(const std::vector< std::set<T> >& in, std::vector< std::set<T> >& out){
  unsigned len = in.size();
  out.clear();
  
  // Write to vectors
  std::vector< std::vector<T> > inout(len);
  for(unsigned i=0; i<len; i++){
    inout[i].resize( in[i].size() );
    unsigned j=0;
    for(typename std::set<T>::const_iterator it=in[i].begin(); it!=in[i].end(); ++it){
      inout[i][j++] = *it;
    }
  }
  
  // Communicate
  allSendRecv(inout);
  
  // Write to sets
  out.resize(len);
  for(unsigned i=0; i<len; i++){
    for(typename std::vector<T>::const_iterator it=inout[i].begin(); it!=inout[i].end(); ++it){
      out[i].insert(*it);
    }
    inout[i].clear();
  }
  
  return;
}

#undef TAG_33
#undef TAG_44
