/* Copyright (C) 2006 Imperial College London and others.

 Please see the AUTHORS file in the main source directory for a full list
 of copyright holders.

 Dr Gerard J Gorman
 Applied Modelling and Computation Group
 Department of Earth Science and Engineering
 Imperial College London

 g.gorman@imperial.ac.uk

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 USA
*/
#include <mpi.h>
#include <assert.h>
#include <vector>
#include <string>

#include "communicationTools.h"
#include "c++debug.h"

using namespace std;

// The purpose of this routine to is coordinate the sending and
// receiving of a vector of packed buffers. The idea is that no-one
// knows in advance how much information that they have to
// receive. For this reason everyone has to communicate something with
// everyone. For this reason, the size of sendbuffer is assumed to be
// the number of processors. Each pair item in the send/recv Buffer
// represents the size of the buffer in char's, and then a vector
// containing the actual buffer.
void allSendRecv(const vector< pair<int, vector<char> > >& sendBuffer, 
		 vector< pair<int, vector<char> > >& recvBuffer){
  unsigned MyRank = MPI::COMM_WORLD.Get_rank();
  unsigned NProcs = sendBuffer.size();
  recvBuffer.clear();
  recvBuffer.resize(NProcs);
  
  // Inform all process how much data they are receiving from every
  // other process. This is implemented using MPI_Alltoall because
  // some implementations of MPI are a bit restrictive on the number
  // of non-blocking communications that can be initiated.
  vector<int> send_count(NProcs), recv_count(NProcs);
  for(size_t p=0;p<(size_t)NProcs;p++){
    send_count[p] = sendBuffer[p].first;
  }
  MPI::COMM_WORLD.Alltoall(&(send_count[0]), 1, MPI::INT,
                           &(recv_count[0]), 1, MPI::INT);
  
  // MPI_Status sendStatus, recvStatus;
  vector<MPI::Request> sendRequest(NProcs, MPI::REQUEST_NULL);
  vector<MPI::Request> recvRequest(NProcs, MPI::REQUEST_NULL);
  MPI::Status recvStatus;
  
  // Send
  for(size_t p=0;p<(size_t)NProcs;p++){
    if(send_count[p]){
      ECHO("MPI Packet of " << sendBuffer[p].first << " being sent to " << p);
      sendRequest[p] = MPI::COMM_WORLD.Isend(sendBuffer[p].second.begin(), send_count[p], MPI::PACKED, p, 1);
    }
  }
  
  // Receive
  for(size_t p=0;p<(size_t)NProcs;p++){
    if(recv_count[p]){
      // Allocate receive space.
      recvBuffer[rank].first = recv_count[p];
      recvBuffer[rank].second.resize(recv_count[p]);
      
      // initiate non-blocking receive
      recvRequest[rank] = MPI::COMM_WORLD.Irecv(recvBuffer[rank].second.begin(), recv_count[p], MPI::PACKED, p, 1);
      
      ECHO("Receiving data from "<<rank);
    }
  }
  
  MPI::Request::Waitall(NProcs, sendRequest.begin()); // Wait all sends
  MPI::Request::Waitall(NProcs, recvRequest.begin()); // Wait all receives

  ECHO("All data received");
 
  return;
}

// In-place send-receive.
void allSendRecv(vector< vector<char> >& inout){
  
  unsigned MyRank = MPI::COMM_WORLD.Get_rank();
  unsigned NProcs = MPI::COMM_WORLD.Get_size();
  assert(NProcs == inout.size());
  
  // Receiving buffer space.
  vector< vector<char> > recvBuffer( NProcs );
  
  // MPI_Status sendStatus, recvStatus;
  vector<MPI::Request> sendRequest( NProcs );
  vector<MPI::Request> recvRequest( NProcs );
  MPI::Status recvStatus;
  
  // Send
  for(unsigned p=0; p < NProcs; p++){
    if(p == MyRank){ // Sending nothing to ourselves.
      sendRequest[ p ] = MPI::REQUEST_NULL;
      continue;
    }
    
    ECHO("MPI Packet of " << inout[p].size() << " being sent to " << p);
    sendRequest[p] = MPI::COMM_WORLD.Isend(inout[p].begin(), inout[p].size(), MPI::PACKED, p, 1);
  }
  
  // Receive
  recvRequest[ MyRank ] = MPI::REQUEST_NULL;    
  for(unsigned p=0; p < NProcs-1; p++){ // -1 because we'll not be receiving from ourselves
    
    // blocking probe for incoming message
    MPI::COMM_WORLD.Probe(MPI::ANY_SOURCE, 1, recvStatus);  
    int mesg_size = recvStatus.Get_count(MPI::PACKED);
    int rank      = recvStatus.Get_source();
    int nbytes;
    {
      MPI::Aint lb;
      MPI::Aint extent;
      MPI::PACKED.Get_extent(lb, extent);
      nbytes    = mesg_size*extent;      
    }

    ECHO("Allocating a "<<nbytes<<" byte buffer for "<<rank);
   
    // Allocate receive space.
    recvBuffer[rank].resize( nbytes );
    
    char *buffer = recvBuffer[rank].begin();
    
    // Initiate non-blocking receive.
    recvRequest[rank] = MPI::COMM_WORLD.Irecv(buffer, mesg_size, MPI::PACKED, rank, 1);
    
    ECHO("Receiving data from "<<rank);
  }
  
  MPI::Request::Waitall(NProcs, sendRequest.begin()); // Wait all sends
  MPI::Request::Waitall(NProcs, recvRequest.begin()); // Wait all receives
  
  sendRequest.clear();
  recvRequest.clear();
  
  for(unsigned i=0; i<NProcs; i++){
    inout[i].swap( recvBuffer[i] );
  }

  ECHO("All data received"); 
  return;
}


void allSendRecv(const vector< vector<unsigned> >& sendBuffer, 
		 vector< vector<unsigned> >& recvBuffer){
  unsigned MyRank = MPI::COMM_WORLD.Get_rank();
  unsigned NProcs = sendBuffer.size();
  
  // Just in case...
  recvBuffer.clear();
  recvBuffer.resize(NProcs);
  
  // MPI_Status sendStatus, recvStatus;
  vector<MPI::Request> sendRequest( NProcs );
  vector<MPI::Request> recvRequest( NProcs );
  MPI::Status recvStatus;
  
  // Set up all the non-blocking sends.
  for(unsigned p=0; p < NProcs; p++){
    if(p == MyRank){ // Sending nothing to ourselves.
      sendRequest[ p ] = MPI::REQUEST_NULL;
      continue;
    }
    
    sendRequest[p] = MPI::COMM_WORLD.Isend(sendBuffer[p].begin(), 
					   sendBuffer[p].size(), 
					   MPI::UNSIGNED, p, 1);
  }
  
  // Set up all the receives. These are set up with the help of an initial probe.
  // Note -1 because we'll not be receiving from ourselves.
  recvRequest[ MyRank ] = MPI::REQUEST_NULL;    
  for(unsigned p=0; p < NProcs-1; p++){ 
    
    // Blocking probe for any incoming message.
    MPI::COMM_WORLD.Probe(MPI::ANY_SOURCE, 1, recvStatus);  
    int mesg_size = recvStatus.Get_count(MPI::UNSIGNED);
    int rank      = recvStatus.Get_source();
    
    // Allocate receive space.
    recvBuffer[rank].resize(mesg_size);
    
    // initiate non-blocking receive
    recvRequest[rank] = MPI::COMM_WORLD.Irecv(recvBuffer[rank].begin(), mesg_size, MPI::UNSIGNED, rank, 1);
  }
  
  // Wait for all sends to complete.
  MPI::Request::Waitall(NProcs, sendRequest.begin());
  // Wait for all receives to complete.
  MPI::Request::Waitall(NProcs, recvRequest.begin());
  
  return;
}













